import yaml
import os
import osmnx as ox
from shapely.geometry import Point, Polygon, LineString, MultiPolygon

from structs.landmarks import Landmark, LandmarkType
from structs.preferences import Preferences, Preference
import constants


SIGHTSEEING = LandmarkType(landmark_type='sightseeing')
NATURE = LandmarkType(landmark_type='nature')
SHOPPING = LandmarkType(landmark_type='shopping')

ox.config(cache_folder=constants.OSM_CACHE_DIR)

# Include the json here
# Create a list of all things to visit given some preferences and a city. Ready for the optimizer
def generate_landmarks(preferences: Preferences, center_coordinates: tuple[float, float]) :
    with constants.AMENITY_SELECTORS_PATH.open('r') as f:
        amenity_selectors = yaml.safe_load(f)
    
    with constants.LANDMARK_PARAMETERS_PATH.open('r') as f:
        # even though we don't use the parameters here, we already load them to avoid unnecessary io operations
        parameters = yaml.safe_load(f)
        max_distance = parameters['city_bbox_side']
    L = []

    # List for sightseeing
    if preferences.sightseeing.score != 0:
        score_func = lambda loc, n_tags: int((count_elements_within_radius(loc, parameters['radius_close_to']) + n_tags * parameters['tag_coeff']) * parameters['church_coeff'])  
        L1 = get_landmarks(amenity_selectors['sightseeing'], SIGHTSEEING, center_coordinates, max_distance, score_func)
        correct_score(L1, preferences.sightseeing)
        L += L1
    
    # List for nature
    if preferences.nature.score != 0:
        score_func = lambda loc, n_tags: int((count_elements_within_radius(loc, parameters['radius_close_to']) + n_tags * parameters['tag_coeff']) * parameters['park_coeff'])  
        L2 = get_landmarks(amenity_selectors['nature'], NATURE, center_coordinates, max_distance, score_func)
        correct_score(L2, preferences.nature)
        L += L2
    
    # List for shopping
    if preferences.shopping.score != 0:
        score_func = lambda loc, n_tags: count_elements_within_radius(loc, parameters['radius_close_to']) + n_tags * parameters['tag_coeff']
        L3 = get_landmarks(amenity_selectors['shopping'], SHOPPING, center_coordinates, max_distance, score_func)
        correct_score(L3, preferences.shopping)
        L += L3

    # remove duplicates
    L = list(set(L))
    print(len(L))
    L_constrained = take_most_important(L, parameters['N_important'])
    print(len(L_constrained))
    return L, L_constrained



# Take the most important landmarks from the list
def take_most_important(landmarks: list[Landmark], n_max: int) -> list[Landmark]:

    landmarks_sorted = sorted(landmarks, key=lambda x: x.attractiveness, reverse=True)
    return landmarks_sorted[:n_max]



# Correct the score of a list of landmarks by taking into account preference settings
def correct_score(L: list[Landmark], preference: Preference) :

    if len(L) == 0 :
        return
    
    if L[0].type != preference.type :
        raise TypeError(f"LandmarkType {preference.type} does not match the type of Landmark {L[0].name}")

    for elem in L :
        elem.attractiveness = int(elem.attractiveness*preference.score/500)     # arbitrary computation


# Function to count elements within a certain radius of a location
def count_elements_within_radius(point: Point, radius: int) -> int:
    
    center_coordinates = (point.x, point.y)
    try:
        landmarks = ox.features_from_point(
            center_point = center_coordinates,
            dist = radius,
            tags = {'building': True} # this is a common tag to give an estimation of the number of elements in the area
        )
        return len(landmarks)
    except ox._errors.InsufficientResponseError:
        return 0    




def get_landmarks(
    amenity_selectors: list[dict],
    landmarktype: LandmarkType,
    center_coordinates: tuple[float, float],
    distance: int,
    score_function: callable
) -> list[Landmark]:
    
    landmarks = ox.features_from_point(
        center_point = center_coordinates,
        dist = distance,
        tags = amenity_selectors
    )

    # cleanup the list
    # remove rows where name is None
    landmarks = landmarks[landmarks['name'].notna()]
    # TODO: remove rows that are part of another building

    ret_landmarks = []
    for element, description in landmarks.iterrows():
        osm_type = element[0]
        osm_id = element[1]
        location = description['geometry']
        n_tags = len(description['nodes']) if type(description['nodes']) == list else 1

        # print(description['nodes'])
        print(description['name'])
        # print(location, type(location))
        if type(location) == Point:
            location = location
        elif type(location) == Polygon or type(location) == MultiPolygon:
            location = location.centroid
        elif type(location) == LineString:
            location = location.interpolate(location.length/2)

        score = score_function(location, n_tags)
        print(score)
        landmark = Landmark(
            name = description['name'],
            type = landmarktype,
            location = (location.x, location.y),
            osm_type = osm_type,
            osm_id = osm_id,
            attractiveness = score,
            must_do = False,
            n_tags = n_tags
        )
        ret_landmarks.append(landmark)

    return ret_landmarks
    # for elem in G.iterrows():
    #     print(elem)
    #     print(elem.name)
    #     print(elem.address)
    #     name = elem.tag('name')                             # Add name
    #     location = (elem.centerLat(), elem.centerLon())     # Add coordinates (lat, lon)

    #     # skip if unprecise location
    #     if name is None or location[0] is None:
    #         continue

    #     # skip if unused
    #     if 'disused:leisure' in elem.tags().keys():
    #         continue
        
    #     # skip if part of another building
    #     if 'building:part' in elem.tags().keys() and elem.tag('building:part') == 'yes':
    #         continue
        
    #     else :
    #         osm_type = elem.type()              # Add type : 'way' or 'relation'
    #         osm_id = elem.id()                  # Add OSM id 
    #         elem_type = landmarktype            # Add the landmark type as 'sightseeing
    #         n_tags = len(elem.tags().keys())    # Add number of tags
            
    #         # Add score of given landmark based on the number of surrounding elements. Penalty for churches as there are A LOT
    #         if amenity == "'amenity'='place_of_worship'" :
    #             score = int((count_elements_within_radius(location, parameters['radius_close_to']) + n_tags*parameters['tag_coeff'] )*parameters['church_coeff'])  
    #         elif amenity == "'leisure'='park'" :
    #             score = int((count_elements_within_radius(location, parameters['radius_close_to']) + n_tags*parameters['tag_coeff'] )*parameters['park_coeff'])  
    #         else :
    #             score = count_elements_within_radius(location, parameters['radius_close_to']) + n_tags*parameters['tag_coeff']

    #         if score is not None :
    #             # Generate the landmark and append it to the list
    #             landmark = Landmark(name=name, type=elem_type, location=location, osm_type=osm_type, osm_id=osm_id, attractiveness=score, must_do=False, n_tags=n_tags)
    #             L.append(landmark)

    # return L


